# std
from typing import Optional

# thirdparty
import numpy as np
from pyvista import PolyData

# local
from . import core
from .fun_tools import cumsum, stagger


def poly2mesh(poly: PolyData) -> core.Mesh:
    poly = poly.triangulate()
    points = poly.points.copy()
    faces = poly.faces.reshape((-1, 4))[:, 1:].copy()
    return (points, faces)


def mesh2poly(mesh: core.Mesh) -> PolyData:
    faces = np.array(mesh[1])
    polys = np.hstack((np.full((faces.shape[0], 1), 3), faces)).ravel()
    return PolyData(mesh[0], polys)


def boolean_function(
    poly1: PolyData,
    poly2: PolyData,
    boolean_operator,
) -> Optional[PolyData]:
    m1 = poly2mesh(poly1)
    m2 = poly2mesh(poly2)
    res, tracks = boolean_operator(m1, m2)
    if res is None:
        raise Exception("core.union failed")
    poly = mesh2poly(res)

    n_faces = len(poly.faces) // 4
    poly["rid"] = np.zeros(n_faces)
    for i, (s0, s1) in enumerate(stagger(cumsum(tracks))):
        poly["rid"][s0:s1] = i
    return poly


def union(poly1: PolyData, poly2: PolyData) -> Optional[PolyData]:
    """
    TODO add documentation
    """
    return boolean_function(poly1, poly2, core.union)


def common(poly1: PolyData, poly2: PolyData) -> Optional[None]:
    """
    TODO add documentation
    """
    return boolean_function(poly1, poly2, core.common)


intersection = common


def diff(poly1: PolyData, poly2: PolyData) -> Optional[None]:
    """
    TODO add documentation
    """
    return boolean_function(poly1, poly2, core.diff)



from typing import Generator, Iterable


def cumsum(a: Iterable) -> Generator:
    """
    the cumalative sum starting from 0
    """
    agg = 0
    yield agg
    for i in a:
        agg += i
        yield agg


def stagger(a: Iterable) -> Generator:
    """
    yield pairs of the newest and second newest
    """
    it = iter(a)

    old = next(it)
    for new in it:
        yield old, new
        old = new



"""
	boolean operation implemented without thirdparty python modules
"""

# std
from typing import List, Tuple, Optional

# local
from .rboolean import pyintersect


Point = Tuple[float, float, float]
Points = List[Point]
Tri = Tuple[int, int, int]
Faces = List[Tri]
Mesh = Tuple[Points, Faces]
Tracks = List[int]

Result = Optional[Tuple[Mesh, Tracks]]


def universal(m1: Mesh, m2: Mesh, flip1: bool, flip2: bool) -> Result:
    """
    calculates boolean operation
    the flip flags determine the type of boolean operation

    returns "None" if the boolean operation fails or the input meshes have no overlap
    """
    # TODO implement exception
    return pyintersect(m1, m2, flip1, flip2)


def union(m1: Mesh, m2: Mesh) -> Result:
    """
    part in m1 or m2
    """
    flip1 = True
    flip2 = True
    return universal(m1, m2, flip1, flip2)


def common(m1: Mesh, m2: Mesh) -> Result:
    """
    part in both m1 and m2
    aka intersection
    """
    flip1 = False
    flip2 = False
    return universal(m1, m2, flip1, flip2)


intersection = common


def diff(m1: Mesh, m2: Mesh) -> Result:
    """
    the part in m1 but not in m2
    """
    flip1 = False
    flip2 = True
    return universal(m1, m2, flip1, flip2)



#!/usr/bin/env python

from setuptools import setup
from setuptools_rust import Binding, RustExtension

setup(
    name="rscad",
    version="0.0.1",
    description="rusty cad utils",
    author="Glenn",
    author_email="gward@python.net",
    rust_extensions=[
        RustExtension(
            "rscad.hello.rhello",
            binding=Binding.PyO3,
            path="./hello/Cargo.toml",
        ),
        RustExtension(
            "rscad.boolean.rboolean",
            binding=Binding.PyO3,
            path="./boolean/Cargo.toml",
        ),
    ],
    packages=[
        "rscad.hello",
        "rscad.boolean",
    ],
    # tell setup that the root python source is inside py folder
    package_dir={
        "rscad.hello": "hello/py",
        "rscad.boolean": "boolean",
    },
    # entry_points={
    #     "console_scripts": ["greet=hello:greet"],
    # },
    zip_safe=False,
)
